@@ -9,28 +9,59 @@ import pytest
 from yarl import URL
 
 import aiohttp
-from aiohttp import helpers
-from aiohttp.client_reqrep import ClientResponse
+from aiohttp import helpers, http
+from aiohttp.client_reqrep import ClientResponse, RequestInfo
 
 
-def test_del(loop):
-    response = ClientResponse('get', URL('http://del-cl-resp.org'))
-    response._post_init(loop)
+@pytest.fixture
+def session():
+    return mock.Mock()
+
+
+@asyncio.coroutine
+def test_http_processing_error(session):
+    loop = mock.Mock()
+    request_info = mock.Mock()
+    response = ClientResponse(
+        'get', URL('http://del-cl-resp.org'), request_info=request_info)
+    response._post_init(loop, session)
+    loop.get_debug = mock.Mock()
+    loop.get_debug.return_value = True
 
     connection = mock.Mock()
-    response._setup_connection(connection)
+    connection.protocol = aiohttp.DataQueue(loop=loop)
+    connection.protocol.set_response_params = mock.Mock()
+    connection.protocol.set_exception(http.HttpProcessingError())
+
+    with pytest.raises(aiohttp.ClientResponseError) as info:
+        yield from response.start(connection)
+
+    assert info.value.request_info is request_info
+
+
+def test_del(session):
+    loop = mock.Mock()
+    response = ClientResponse('get', URL('http://del-cl-resp.org'))
+    response._post_init(loop, session)
+    loop.get_debug = mock.Mock()
+    loop.get_debug.return_value = True
+
+    connection = mock.Mock()
+    response._closed = False
+    response._connection = connection
     loop.set_exception_handler(lambda loop, ctx: None)
 
     with pytest.warns(ResourceWarning):
         del response
         gc.collect()
 
-    connection.close.assert_called_with()
+    connection.release.assert_called_with()
 
 
-def test_close(loop):
+def test_close(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
+    response._closed = False
     response._connection = mock.Mock()
     response.close()
     assert response.connection is None
@@ -38,25 +69,25 @@ def test_close(loop):
     response.close()
 
 
-def test_wait_for_100_1(loop):
+def test_wait_for_100_1(loop, session):
     response = ClientResponse(
         'get', URL('http://python.org'), continue100=object())
-    response._post_init(loop)
+    response._post_init(loop, session)
     assert response._continue is not None
     response.close()
 
 
-def test_wait_for_100_2(loop):
+def test_wait_for_100_2(loop, session):
     response = ClientResponse(
         'get', URL('http://python.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
     assert response._continue is None
     response.close()
 
 
-def test_repr(loop):
+def test_repr(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
     response.status = 200
     response.reason = 'Ok'
     assert '<ClientResponse(http://def-cl-resp.org) [200 Ok]>'\
@@ -76,10 +107,16 @@ def test_repr_non_ascii_reason():
         in repr(response)
 
 
+def test_url_obj_deprecated():
+    response = ClientResponse('get', URL('http://fake-host.org/'))
+    with pytest.warns(DeprecationWarning):
+        response.url_obj
+
+
 @asyncio.coroutine
-def test_read_and_release_connection(loop):
+def test_read_and_release_connection(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -94,9 +131,9 @@ def test_read_and_release_connection(loop):
 
 
 @asyncio.coroutine
-def test_read_and_release_connection_with_error(loop):
+def test_read_and_release_connection_with_error(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
     content = response.content = mock.Mock()
     content.read.return_value = helpers.create_future(loop)
     content.read.return_value.set_exception(ValueError)
@@ -107,22 +144,77 @@ def test_read_and_release_connection_with_error(loop):
 
 
 @asyncio.coroutine
-def test_release(loop):
+def test_release(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
     fut = helpers.create_future(loop)
     fut.set_result(b'')
     content = response.content = mock.Mock()
     content.readany.return_value = fut
 
-    yield from response.release()
+    response.release()
     assert response._connection is None
 
 
 @asyncio.coroutine
-def test_text(loop):
+def test_release_on_del(loop, session):
+    connection = mock.Mock()
+    connection.protocol.upgraded = False
+
+    def run(conn):
+        response = ClientResponse('get', URL('http://def-cl-resp.org'))
+        response._post_init(loop, session)
+        response._closed = False
+        response._connection = conn
+
+    run(connection)
+
+    assert connection.release.called
+
+
+@asyncio.coroutine
+def test_response_eof(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
+    response._closed = False
+    conn = response._connection = mock.Mock()
+    conn.protocol.upgraded = False
+
+    response._response_eof()
+    assert conn.release.called
+    assert response._connection is None
+
+
+@asyncio.coroutine
+def test_response_eof_upgraded(loop, session):
+    response = ClientResponse('get', URL('http://def-cl-resp.org'))
+    response._post_init(loop, session)
+
+    conn = response._connection = mock.Mock()
+    conn.protocol.upgraded = True
+
+    response._response_eof()
+    assert not conn.release.called
+    assert response._connection is conn
+
+
+@asyncio.coroutine
+def test_response_eof_after_connection_detach(loop, session):
+    response = ClientResponse('get', URL('http://def-cl-resp.org'))
+    response._post_init(loop, session)
+    response._closed = False
+    conn = response._connection = mock.Mock()
+    conn.protocol = None
+
+    response._response_eof()
+    assert conn.release.called
+    assert response._connection is None
+
+
+@asyncio.coroutine
+def test_text(loop, session):
+    response = ClientResponse('get', URL('http://def-cl-resp.org'))
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -140,9 +232,32 @@ def test_text(loop):
 
 
 @asyncio.coroutine
-def test_text_custom_encoding(loop):
+def test_text_bad_encoding(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
+
+    def side_effect(*args, **kwargs):
+        fut = helpers.create_future(loop)
+        fut.set_result('{"тестkey": "пройденvalue"}'.encode('cp1251'))
+        return fut
+
+    # lie about the encoding
+    response.headers = {
+        'Content-Type': 'application/json;charset=utf-8'}
+    content = response.content = mock.Mock()
+    content.read.side_effect = side_effect
+    with pytest.raises(UnicodeDecodeError):
+        yield from response.text()
+    # only the valid utf-8 characters will be returned
+    res = yield from response.text(errors='ignore')
+    assert res == '{"key": "value"}'
+    assert response._connection is None
+
+
+@asyncio.coroutine
+def test_text_custom_encoding(loop, session):
+    response = ClientResponse('get', URL('http://def-cl-resp.org'))
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -162,9 +277,9 @@ def test_text_custom_encoding(loop):
 
 
 @asyncio.coroutine
-def test_text_detect_encoding(loop):
+def test_text_detect_encoding(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -182,9 +297,9 @@ def test_text_detect_encoding(loop):
 
 
 @asyncio.coroutine
-def test_text_after_read(loop):
+def test_text_after_read(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -202,9 +317,9 @@ def test_text_after_read(loop):
 
 
 @asyncio.coroutine
-def test_json(loop):
+def test_json(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -222,9 +337,9 @@ def test_json(loop):
 
 
 @asyncio.coroutine
-def test_json_custom_loader(loop):
+def test_json_custom_loader(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
     response.headers = {
         'Content-Type': 'application/json;charset=cp1251'}
     response._content = b'data'
@@ -237,26 +352,26 @@ def test_json_custom_loader(loop):
 
 
 @asyncio.coroutine
-def test_json_no_content(loop):
+def test_json_no_content(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
     response.headers = {
         'Content-Type': 'data/octet-stream'}
     response._content = b''
 
-    with mock.patch('aiohttp.client_reqrep.client_logger') as m_log:
-        res = yield from response.json()
+    with pytest.raises(aiohttp.ClientResponseError) as info:
+        yield from response.json()
 
+    assert info.value.request_info == response.request_info
+
+    res = yield from response.json(content_type=None)
     assert res is None
-    m_log.warning.assert_called_with(
-        'Attempt to decode JSON with unexpected mimetype: %s',
-        'data/octet-stream')
 
 
 @asyncio.coroutine
-def test_json_override_encoding(loop):
+def test_json_override_encoding(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
 
     def side_effect(*args, **kwargs):
         fut = helpers.create_future(loop)
@@ -275,19 +390,20 @@ def test_json_override_encoding(loop):
     assert not response._get_encoding.called
 
 
-def test_override_flow_control(loop):
+@pytest.mark.xfail
+def test_override_flow_control(loop, session):
     class MyResponse(ClientResponse):
         flow_control_class = aiohttp.StreamReader
     response = MyResponse('get', URL('http://my-cl-resp.org'))
-    response._post_init(loop)
-    response._setup_connection(mock.Mock())
+    response._post_init(loop, session)
+    response._connection = mock.Mock()
     assert isinstance(response.content, aiohttp.StreamReader)
     response.close()
 
 
-def test_get_encoding_unknown(loop):
+def test_get_encoding_unknown(loop, session):
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
-    response._post_init(loop)
+    response._post_init(loop, session)
 
     response.headers = {'Content-Type': 'application/json'}
     with mock.patch('aiohttp.client_reqrep.chardet') as m_chardet:
@@ -306,7 +422,7 @@ def test_raise_for_status_4xx():
     response = ClientResponse('get', URL('http://def-cl-resp.org'))
     response.status = 409
     response.reason = 'CONFLICT'
-    with pytest.raises(aiohttp.HttpProcessingError) as cm:
+    with pytest.raises(aiohttp.ClientResponseError) as cm:
         response.raise_for_status()
     assert str(cm.value.code) == '409'
     assert str(cm.value.message) == "CONFLICT"
@@ -314,8 +430,7 @@ def test_raise_for_status_4xx():
 
 def test_resp_host():
     response = ClientResponse('get', URL('http://del-cl-resp.org'))
-    with pytest.warns(DeprecationWarning):
-        assert 'del-cl-resp.org' == response.host
+    assert 'del-cl-resp.org' == response.host
 
 
 def test_content_type():
@@ -351,3 +466,104 @@ def test_charset_no_charset():
     response.headers = {'Content-Type': 'application/json'}
 
     assert response.charset is None
+
+
+def test_response_request_info():
+    url = 'http://def-cl-resp.org'
+    headers = {'Content-Type': 'application/json;charset=cp1251'}
+    response = ClientResponse(
+        'get', URL(url),
+        request_info=RequestInfo(
+            url,
+            'get',
+            headers
+        )
+    )
+    assert url == response.request_info.url
+    assert 'get' == response.request_info.method
+    assert headers == response.request_info.headers
+
+
+def test_response_request_info_empty():
+    url = 'http://def-cl-resp.org'
+    response = ClientResponse(
+        'get', URL(url),
+    )
+    assert response.request_info is None
+
+
+def test_request_info_in_exception():
+    url = 'http://def-cl-resp.org'
+    headers = {'Content-Type': 'application/json;charset=cp1251'}
+    response = ClientResponse(
+        'get',
+        URL(url),
+        request_info=RequestInfo(
+            url,
+            'get',
+            headers
+        )
+    )
+    response.status = 409
+    response.reason = 'CONFLICT'
+    with pytest.raises(aiohttp.ClientResponseError) as cm:
+        response.raise_for_status()
+    assert cm.value.request_info == response.request_info
+
+
+def test_no_redirect_history_in_exception():
+    url = 'http://def-cl-resp.org'
+    headers = {'Content-Type': 'application/json;charset=cp1251'}
+    response = ClientResponse(
+        'get',
+        URL(url),
+        request_info=RequestInfo(
+            url,
+            'get',
+            headers
+        )
+    )
+    response.status = 409
+    response.reason = 'CONFLICT'
+    with pytest.raises(aiohttp.ClientResponseError) as cm:
+        response.raise_for_status()
+    assert () == cm.value.history
+
+
+def test_redirect_history_in_exception():
+    hist_url = 'http://def-cl-resp.org'
+    url = 'http://def-cl-resp.org/index.htm'
+    hist_headers = {'Content-Type': 'application/json;charset=cp1251',
+                    'Location': url
+                    }
+    headers = {'Content-Type': 'application/json;charset=cp1251'}
+    response = ClientResponse(
+        'get',
+        URL(url),
+        request_info=RequestInfo(
+            url,
+            'get',
+            headers
+        )
+    )
+    response.status = 409
+    response.reason = 'CONFLICT'
+
+    hist_response = ClientResponse(
+        'get',
+        URL(hist_url),
+        request_info=RequestInfo(
+            url,
+            'get',
+            headers
+        )
+    )
+
+    hist_response.headers = hist_headers
+    hist_response.status = 301
+    hist_response.reason = 'REDIRECT'
+
+    response._history = [hist_response]
+    with pytest.raises(aiohttp.ClientResponseError) as cm:
+        response.raise_for_status()
+    assert [hist_response] == cm.value.history
