@@ -12,6 +12,6 @@ async def test_render_ok():
             return resp
 
     request = mock.Mock()
-    request.method = 'GET'
+    request._method = 'GET'
     resp2 = await MyView(request)
     assert resp is resp2
