@@ -1,4 +1,5 @@
 import asyncio
+import io
 import json
 import pathlib
 import zlib
@@ -8,8 +9,9 @@ import pytest
 from multidict import MultiDict
 from yarl import URL
 
-from aiohttp import FormData, multipart, web
-from aiohttp.protocol import HttpVersion, HttpVersion10, HttpVersion11
+import aiohttp
+from aiohttp import FormData, HttpVersion10, HttpVersion11, multipart, web
+
 
 try:
     import ssl
@@ -17,6 +19,16 @@ except:
     ssl = False
 
 
+@pytest.fixture
+def here():
+    return pathlib.Path(__file__).parent
+
+
+@pytest.fixture
+def fname(here):
+    return here / 'sample.key'
+
+
 @asyncio.coroutine
 def test_simple_get(loop, test_client):
 
@@ -26,7 +38,26 @@ def test_simple_get(loop, test_client):
         assert b'' == body
         return web.Response(body=b'OK')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    txt = yield from resp.text()
+    assert 'OK' == txt
+
+
+@asyncio.coroutine
+def test_simple_get_with_text(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        body = yield from request.read()
+        assert b'' == body
+        return web.Response(text='OK', headers={'content-type': 'text/plain'})
+
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -44,7 +75,7 @@ def test_handler_returns_not_response(loop, test_server, test_client):
     def handler(request):
         return 'abc'
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     server = yield from test_server(app, logger=logger)
     client = yield from test_client(server)
@@ -62,30 +93,49 @@ def test_head_returns_empty_body(loop, test_client):
     def handler(request):
         return web.Response(body=b'test')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_head('/', handler)
-    client = yield from test_client(app)
+    client = yield from test_client(app, version=HttpVersion11)
 
-    resp = yield from client.head('/', version=HttpVersion11)
+    resp = yield from client.head('/')
     assert 200 == resp.status
     txt = yield from resp.text()
     assert '' == txt
 
 
+@asyncio.coroutine
+def test_response_before_complete(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(body=b'OK')
+
+    app = web.Application()
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+
+    data = b'0' * 1024 * 1024
+
+    resp = yield from client.post('/', data=data)
+    assert 200 == resp.status
+    text = yield from resp.text()
+    assert 'OK' == text
+
+
 @asyncio.coroutine
 def test_post_form(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
         data = yield from request.post()
-        assert {'a': '1', 'b': '2'} == data
+        assert {'a': '1', 'b': '2', 'c': ''} == data
         return web.Response(body=b'OK')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
-    resp = yield from client.post('/', data={'a': 1, 'b': 2})
+    resp = yield from client.post('/', data={'a': 1, 'b': 2, 'c': ''})
     assert 200 == resp.status
     txt = yield from resp.text()
     assert 'OK' == txt
@@ -102,7 +152,7 @@ def test_post_text(loop, test_client):
         assert data == data2
         return web.Response(text=data)
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -123,15 +173,12 @@ def test_post_json(loop, test_client):
         assert dct == data
         data2 = yield from request.json(loads=json.loads)
         assert data == data2
-        with pytest.warns(DeprecationWarning):
-            data3 = yield from request.json(loader=json.loads)
-        assert data == data3
         resp = web.Response()
         resp.content_type = 'application/json'
         resp.body = json.dumps(data).encode('utf8')
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -169,7 +216,7 @@ def test_multipart(loop, test_client):
         resp.body = b''
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -201,7 +248,7 @@ def test_multipart_content_transfer_encoding(loop, test_client):
         resp.body = b''
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -217,7 +264,7 @@ def test_render_redirect(loop, test_client):
     def handler(request):
         raise web.HTTPMovedPermanently(location='/path')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -242,8 +289,7 @@ def test_post_single_file(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
-        with pytest.warns(DeprecationWarning):
-            data = yield from request.post()
+        data = yield from request.post()
         assert ['sample.crt'] == list(data.keys())
         for fs in data.values():
             check_file(fs)
@@ -251,7 +297,7 @@ def test_post_single_file(loop, test_client):
         resp = web.Response(body=b'OK')
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -280,7 +326,7 @@ def test_files_upload_with_same_key(loop, test_client):
         resp = web.Response(body=b'OK')
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -317,7 +363,7 @@ def test_post_files(loop, test_client):
         resp = web.Response(body=b'OK')
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -337,7 +383,7 @@ def test_release_post_data(loop, test_client):
         assert chunk == b''
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -353,7 +399,7 @@ def test_POST_DATA_with_content_transfer_encoding(loop, test_client):
         assert b'123' == data['name']
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -374,7 +420,7 @@ def test_post_form_with_duplicate_keys(loop, test_client):
         assert [('a', '1'), ('a', '2')] == lst
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -383,7 +429,7 @@ def test_post_form_with_duplicate_keys(loop, test_client):
 
 
 def test_repr_for_application(loop):
-    app = web.Application(loop=loop)
+    app = web.Application()
     assert "<Application 0x{:x}>".format(id(app)) == repr(app)
 
 
@@ -406,7 +452,7 @@ def test_expect_default_handler_unknown(loop, test_client):
         pytest.xfail('Handler should not proceed to this point in case of '
                      'unknown Expect header')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -426,7 +472,7 @@ def test_100_continue(loop, test_client):
     form.add_field('name', b'123',
                    content_transfer_encoding='base64')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -456,7 +502,7 @@ def test_100_continue_custom(loop, test_client):
     form.add_field('name', b'123',
                    content_transfer_encoding='base64')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler, expect_handler=expect_handler)
     client = yield from test_client(app)
 
@@ -480,13 +526,13 @@ def test_100_continue_custom_response(loop, test_client):
             if auth_err:
                 return web.HTTPForbidden()
 
-            request.transport.write(b"HTTP/1.1 100 Continue\r\n\r\n")
+            request.writer.write(b"HTTP/1.1 100 Continue\r\n\r\n")
 
     form = FormData()
     form.add_field('name', b'123',
                    content_transfer_encoding='base64')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler, expect_handler=expect_handler)
     client = yield from test_client(app)
 
@@ -502,7 +548,7 @@ def test_100_continue_custom_response(loop, test_client):
 @asyncio.coroutine
 def test_100_continue_for_not_found(loop, test_client):
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     client = yield from test_client(app)
 
     resp = yield from client.post('/not_found', data='data', expect100=True)
@@ -516,7 +562,7 @@ def test_100_continue_for_not_allowed(loop, test_client):
     def handler(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -531,16 +577,17 @@ def test_http11_keep_alive_default(loop, test_client):
     def handler(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
-    client = yield from test_client(app)
+    client = yield from test_client(app, version=HttpVersion11)
 
-    resp = yield from client.get('/', version=HttpVersion11)
+    resp = yield from client.get('/')
     assert 200 == resp.status
     assert resp.version == HttpVersion11
     assert 'Connection' not in resp.headers
 
 
+@pytest.mark.xfail
 @asyncio.coroutine
 def test_http10_keep_alive_default(loop, test_client):
 
@@ -548,36 +595,16 @@ def test_http10_keep_alive_default(loop, test_client):
     def handler(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
-    client = yield from test_client(app)
+    client = yield from test_client(app, version=HttpVersion10)
 
-    resp = yield from client.get('/', version=HttpVersion10)
+    resp = yield from client.get('/')
     assert 200 == resp.status
     assert resp.version == HttpVersion10
     assert resp.headers['Connection'] == 'keep-alive'
 
 
-@asyncio.coroutine
-def test_http09_keep_alive_default(loop, test_client):
-
-    @asyncio.coroutine
-    def handler(request):
-        yield from request.read()
-        return web.Response()
-
-    app = web.Application(loop=loop)
-    app.router.add_get('/', handler)
-    client = yield from test_client(app)
-
-    headers = {'Connection': 'keep-alive'}  # should be ignored
-    resp = yield from client.get('/', version=HttpVersion(0, 9),
-                                 headers=headers)
-    assert 200 == resp.status
-    assert resp.version == HttpVersion(0, 9)
-    assert 'Connection' not in resp.headers
-
-
 @asyncio.coroutine
 def test_http10_keep_alive_with_headers_close(loop, test_client):
 
@@ -586,13 +613,12 @@ def test_http10_keep_alive_with_headers_close(loop, test_client):
         yield from request.read()
         return web.Response(body=b'OK')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
-    client = yield from test_client(app)
+    client = yield from test_client(app, version=HttpVersion10)
 
     headers = {'Connection': 'close'}
-    resp = yield from client.get('/', version=HttpVersion10,
-                                 headers=headers)
+    resp = yield from client.get('/', headers=headers)
     assert 200 == resp.status
     assert resp.version == HttpVersion10
     assert 'Connection' not in resp.headers
@@ -606,13 +632,12 @@ def test_http10_keep_alive_with_headers(loop, test_client):
         yield from request.read()
         return web.Response(body=b'OK')
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
-    client = yield from test_client(app)
+    client = yield from test_client(app, version=HttpVersion10)
 
     headers = {'Connection': 'keep-alive'}
-    resp = yield from client.get('/', version=HttpVersion10,
-                                 headers=headers)
+    resp = yield from client.get('/', headers=headers)
     assert 200 == resp.status
     assert resp.version == HttpVersion10
     assert resp.headers['Connection'] == 'keep-alive'
@@ -622,7 +647,7 @@ def test_http10_keep_alive_with_headers(loop, test_client):
 def test_upload_file(loop, test_client):
 
     here = pathlib.Path(__file__).parent
-    fname = here / 'software_development_in_picture.jpg'
+    fname = here / 'aiohttp.png'
     with fname.open('rb') as f:
         data = f.read()
 
@@ -633,7 +658,7 @@ def test_upload_file(loop, test_client):
         assert data == raw_data
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -644,7 +669,7 @@ def test_upload_file(loop, test_client):
 @asyncio.coroutine
 def test_upload_file_object(loop, test_client):
     here = pathlib.Path(__file__).parent
-    fname = here / 'software_development_in_picture.jpg'
+    fname = here / 'aiohttp.png'
     with fname.open('rb') as f:
         data = f.read()
 
@@ -655,7 +680,7 @@ def test_upload_file_object(loop, test_client):
         assert data == raw_data
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -672,7 +697,7 @@ def test_empty_content_for_query_without_body(loop, test_client):
         assert not request.has_body
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -689,7 +714,7 @@ def test_empty_content_for_query_with_body(loop, test_client):
         body = yield from request.read()
         return web.Response(body=body)
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_post('/', handler)
     client = yield from test_client(app)
 
@@ -702,11 +727,11 @@ def test_get_with_empty_arg(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
-        assert 'arg' in request.GET
-        assert '' == request.GET['arg']
+        assert 'arg' in request.query
+        assert '' == request.query['arg']
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -721,7 +746,7 @@ def test_large_header(loop, test_client):
     def handler(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -737,7 +762,7 @@ def test_large_header_allowed(loop, test_client, test_server):
     def handler(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     server = yield from test_server(app, max_field_size=81920)
     client = yield from test_client(server)
@@ -751,11 +776,11 @@ def test_large_header_allowed(loop, test_client, test_server):
 def test_get_with_empty_arg_with_equal(loop, test_client):
     @asyncio.coroutine
     def handler(request):
-        assert 'arg' in request.GET
-        assert '' == request.GET['arg']
+        assert 'arg' in request.query
+        assert '' == request.query['arg']
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -763,16 +788,196 @@ def test_get_with_empty_arg_with_equal(loop, test_client):
     assert 200 == resp.status
 
 
-@pytest.mark.xfail  # and had never worked
 @asyncio.coroutine
-def test_response_with_precompressed_body(loop, test_client):
+def test_response_with_streamer(loop, test_client, fname):
+
+    with fname.open('rb') as f:
+        data = f.read()
+
+    data_size = len(data)
+
+    @aiohttp.streamer
+    def stream(writer, f_name):
+        with f_name.open('rb') as f:
+            data = f.read(100)
+            while data:
+                yield from writer.write(data)
+                data = f.read(100)
+
+    @asyncio.coroutine
+    def handler(request):
+        headers = {'Content-Length': str(data_size)}
+        return web.Response(body=stream(fname), headers=headers)
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    resp_data = yield from resp.read()
+    assert resp_data == data
+    assert resp.headers.get('Content-Length') == str(len(resp_data))
+
+
+@asyncio.coroutine
+def test_response_with_streamer_no_params(loop, test_client, fname):
+
+    with fname.open('rb') as f:
+        data = f.read()
+
+    data_size = len(data)
+
+    @aiohttp.streamer
+    def stream(writer):
+        with fname.open('rb') as f:
+            data = f.read(100)
+            while data:
+                yield from writer.write(data)
+                data = f.read(100)
+
+    @asyncio.coroutine
+    def handler(request):
+        headers = {'Content-Length': str(data_size)}
+        return web.Response(body=stream, headers=headers)
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    resp_data = yield from resp.read()
+    assert resp_data == data
+    assert resp.headers.get('Content-Length') == str(len(resp_data))
+
+
+@asyncio.coroutine
+def test_response_with_file(loop, test_client, fname):
+
+    with fname.open('rb') as f:
+        data = f.read()
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(body=fname.open('rb'))
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    resp_data = yield from resp.read()
+    assert resp_data == data
+    assert resp.headers.get('Content-Type') in (
+        'application/octet-stream', 'application/pgp-keys')
+    assert resp.headers.get('Content-Length') == str(len(resp_data))
+    assert (resp.headers.get('Content-Disposition') ==
+            'attachment; filename="sample.key"; filename*=utf-8\'\'sample.key')
+
+
+@asyncio.coroutine
+def test_response_with_file_ctype(loop, test_client, fname):
+
+    with fname.open('rb') as f:
+        data = f.read()
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(
+            body=fname.open('rb'), headers={'content-type': 'text/binary'})
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    resp_data = yield from resp.read()
+    assert resp_data == data
+    assert resp.headers.get('Content-Type') == 'text/binary'
+    assert resp.headers.get('Content-Length') == str(len(resp_data))
+    assert (resp.headers.get('Content-Disposition') ==
+            'attachment; filename="sample.key"; filename*=utf-8\'\'sample.key')
+
+
+@asyncio.coroutine
+def test_response_with_payload_disp(loop, test_client, fname):
+
+    with fname.open('rb') as f:
+        data = f.read()
+
+    @asyncio.coroutine
+    def handler(request):
+        pl = aiohttp.get_payload(fname.open('rb'))
+        pl.set_content_disposition('inline', filename='test.txt')
+        return web.Response(
+            body=pl, headers={'content-type': 'text/binary'})
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    resp_data = yield from resp.read()
+    assert resp_data == data
+    assert resp.headers.get('Content-Type') == 'text/binary'
+    assert resp.headers.get('Content-Length') == str(len(resp_data))
+    assert (resp.headers.get('Content-Disposition') ==
+            'inline; filename="test.txt"; filename*=utf-8\'\'test.txt')
+
+
+@asyncio.coroutine
+def test_response_with_payload_stringio(loop, test_client, fname):
+
+    @asyncio.coroutine
+    def handler(request):
+        return web.Response(body=io.StringIO('test'))
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    resp_data = yield from resp.read()
+    assert resp_data == b'test'
+
+
+@asyncio.coroutine
+def test_response_with_precompressed_body_gzip(loop, test_client):
+
     @asyncio.coroutine
     def handler(request):
         headers = {'Content-Encoding': 'gzip'}
-        deflated_data = zlib.compress(b'mydata')
-        return web.Response(body=deflated_data, headers=headers)
+        zcomp = zlib.compressobj(wbits=16 + zlib.MAX_WBITS)
+        data = zcomp.compress(b'mydata') + zcomp.flush()
+        return web.Response(body=data, headers=headers)
 
-    app = web.Application(loop=loop)
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get('/')
+    assert 200 == resp.status
+    data = yield from resp.read()
+    assert b'mydata' == data
+    assert resp.headers.get('Content-Encoding') == 'gzip'
+
+
+@asyncio.coroutine
+def test_response_with_precompressed_body_deflate(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        headers = {'Content-Encoding': 'deflate'}
+        zcomp = zlib.compressobj(wbits=-zlib.MAX_WBITS)
+        data = zcomp.compress(b'mydata') + zcomp.flush()
+        return web.Response(body=data, headers=headers)
+
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -783,6 +988,27 @@ def test_response_with_precompressed_body(loop, test_client):
     assert resp.headers.get('Content-Encoding') == 'deflate'
 
 
+@asyncio.coroutine
+def test_bad_request_payload(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        assert request.method == 'GET'
+
+        with pytest.raises(aiohttp.web.RequestPayloadError):
+            yield from request.content.read()
+
+        return web.Response()
+
+    app = web.Application()
+    app.router.add_get('/', handler)
+    client = yield from test_client(app)
+
+    resp = yield from client.get(
+        '/', data=b'test', headers={'content-encoding': 'gzip'})
+    assert 200 == resp.status
+
+
 @asyncio.coroutine
 def test_stream_response_multiple_chunks(loop, test_client):
     @asyncio.coroutine
@@ -795,7 +1021,7 @@ def test_stream_response_multiple_chunks(loop, test_client):
         resp.write(b'z')
         return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -808,7 +1034,7 @@ def test_stream_response_multiple_chunks(loop, test_client):
 @asyncio.coroutine
 def test_start_without_routes(loop, test_client):
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     client = yield from test_client(app)
 
     resp = yield from client.get('/')
@@ -822,7 +1048,7 @@ def test_requests_count(loop, test_client):
     def handler(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
     assert client.server.handler.requests_count == 0
@@ -851,7 +1077,7 @@ def test_redirect_url(loop, test_client):
     def redirected(request):
         return web.Response()
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/redirector', redirector)
     app.router.add_get('/redirected', redirected)
 
@@ -866,10 +1092,10 @@ def test_simple_subapp(loop, test_client):
     def handler(request):
         return web.Response(text="OK")
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
-    app.router.add_subapp('/path', subapp)
+    app.add_subapp('/path', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/to')
@@ -889,18 +1115,18 @@ def test_subapp_reverse_url(loop, test_client):
     def handler2(request):
         return web.Response(text="OK")
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
     subapp.router.add_get('/final', handler2, name='name')
-    app.router.add_subapp('/path', subapp)
+    app.add_subapp('/path', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/to')
     assert resp.status == 200
     txt = yield from resp.text()
     assert 'OK' == txt
-    assert resp.url_obj.path == '/path/final'
+    assert resp.url.path == '/path/final'
 
 
 @asyncio.coroutine
@@ -914,39 +1140,39 @@ def test_subapp_reverse_variable_url(loop, test_client):
     def handler2(request):
         return web.Response(text="OK")
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
     subapp.router.add_get('/{part}', handler2, name='name')
-    app.router.add_subapp('/path', subapp)
+    app.add_subapp('/path', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/to')
     assert resp.status == 200
     txt = yield from resp.text()
     assert 'OK' == txt
-    assert resp.url_obj.path == '/path/final'
+    assert resp.url.path == '/path/final'
 
 
 @asyncio.coroutine
 def test_subapp_reverse_static_url(loop, test_client):
-    fname = 'software_development_in_picture.jpg'
+    fname = 'aiohttp.png'
 
     @asyncio.coroutine
     def handler(request):
         return web.HTTPMovedPermanently(
             location=subapp.router['name'].url_for(filename=fname))
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
     here = pathlib.Path(__file__).parent
     subapp.router.add_static('/static', here, name='name')
-    app.router.add_subapp('/path', subapp)
+    app.add_subapp('/path', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/to')
-    assert resp.url_obj.path == '/path/static/' + fname
+    assert resp.url.path == '/path/static/' + fname
     assert resp.status == 200
     body = yield from resp.read()
     with (here / fname).open('rb') as f:
@@ -960,10 +1186,10 @@ def test_subapp_app(loop, test_client):
         assert request.app is subapp
         return web.HTTPOk(text='OK')
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
-    app.router.add_subapp('/path/', subapp)
+    app.add_subapp('/path/', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/to')
@@ -978,10 +1204,10 @@ def test_subapp_not_found(loop, test_client):
     def handler(request):
         return web.HTTPOk(text='OK')
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
-    app.router.add_subapp('/path/', subapp)
+    app.add_subapp('/path/', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/other')
@@ -994,10 +1220,10 @@ def test_subapp_not_found2(loop, test_client):
     def handler(request):
         return web.HTTPOk(text='OK')
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
-    app.router.add_subapp('/path/', subapp)
+    app.add_subapp('/path/', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/invalid/other')
@@ -1010,15 +1236,15 @@ def test_subapp_not_allowed(loop, test_client):
     def handler(request):
         return web.HTTPOk(text='OK')
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
-    app.router.add_subapp('/path/', subapp)
+    app.add_subapp('/path/', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.post('/path/to')
     assert resp.status == 405
-    assert resp.headers['Allow'] == 'GET'
+    assert resp.headers['Allow'] == 'GET,HEAD'
 
 
 @asyncio.coroutine
@@ -1028,10 +1254,10 @@ def test_subapp_cannot_add_app_in_handler(loop, test_client):
         request.match_info.add_app(app)
         return web.HTTPOk(text='OK')
 
-    app = web.Application(loop=loop)
-    subapp = web.Application(loop=loop)
+    app = web.Application()
+    subapp = web.Application()
     subapp.router.add_get('/to', handler)
-    app.router.add_subapp('/path/', subapp)
+    app.add_subapp('/path/', subapp)
 
     client = yield from test_client(app)
     resp = yield from client.get('/path/to')
@@ -1058,12 +1284,12 @@ def test_subapp_middlewares(loop, test_client):
             return resp
         return middleware
 
-    app = web.Application(loop=loop, middlewares=[middleware_factory])
-    subapp1 = web.Application(loop=loop, middlewares=[middleware_factory])
-    subapp2 = web.Application(loop=loop, middlewares=[middleware_factory])
+    app = web.Application(middlewares=[middleware_factory])
+    subapp1 = web.Application(middlewares=[middleware_factory])
+    subapp2 = web.Application(middlewares=[middleware_factory])
     subapp2.router.add_get('/to', handler)
-    subapp1.router.add_subapp('/b/', subapp2)
-    app.router.add_subapp('/a/', subapp1)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
 
     client = yield from test_client(app)
     resp = yield from client.get('/a/b/to')
@@ -1088,15 +1314,15 @@ def test_subapp_on_response_prepare(loop, test_client):
 
         return on_response
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.on_response_prepare.append(make_signal(app))
-    subapp1 = web.Application(loop=loop)
+    subapp1 = web.Application()
     subapp1.on_response_prepare.append(make_signal(subapp1))
-    subapp2 = web.Application(loop=loop)
+    subapp2 = web.Application()
     subapp2.on_response_prepare.append(make_signal(subapp2))
     subapp2.router.add_get('/to', handler)
-    subapp1.router.add_subapp('/b/', subapp2)
-    app.router.add_subapp('/a/', subapp1)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
 
     client = yield from test_client(app)
     resp = yield from client.get('/a/b/to')
@@ -1112,14 +1338,14 @@ def test_subapp_on_startup(loop, test_server):
     def on_signal(app):
         order.append(app)
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.on_startup.append(on_signal)
-    subapp1 = web.Application(loop=loop)
+    subapp1 = web.Application()
     subapp1.on_startup.append(on_signal)
-    subapp2 = web.Application(loop=loop)
+    subapp2 = web.Application()
     subapp2.on_startup.append(on_signal)
-    subapp1.router.add_subapp('/b/', subapp2)
-    app.router.add_subapp('/a/', subapp1)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
 
     yield from test_server(app)
 
@@ -1133,14 +1359,14 @@ def test_subapp_on_shutdown(loop, test_server):
     def on_signal(app):
         order.append(app)
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.on_shutdown.append(on_signal)
-    subapp1 = web.Application(loop=loop)
+    subapp1 = web.Application()
     subapp1.on_shutdown.append(on_signal)
-    subapp2 = web.Application(loop=loop)
+    subapp2 = web.Application()
     subapp2.on_shutdown.append(on_signal)
-    subapp1.router.add_subapp('/b/', subapp2)
-    app.router.add_subapp('/a/', subapp1)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
 
     server = yield from test_server(app)
     yield from server.close()
@@ -1156,14 +1382,14 @@ def test_subapp_on_cleanup(loop, test_server):
     def on_signal(app):
         order.append(app)
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.on_cleanup.append(on_signal)
-    subapp1 = web.Application(loop=loop)
+    subapp1 = web.Application()
     subapp1.on_cleanup.append(on_signal)
-    subapp2 = web.Application(loop=loop)
+    subapp2 = web.Application()
     subapp2.on_cleanup.append(on_signal)
-    subapp1.router.add_subapp('/b/', subapp2)
-    app.router.add_subapp('/a/', subapp1)
+    subapp1.add_subapp('/b/', subapp2)
+    app.add_subapp('/a/', subapp1)
 
     server = yield from test_server(app)
     yield from server.close()
@@ -1178,7 +1404,7 @@ def test_custom_date_header(loop, test_client):
     def handler(request):
         return web.Response(headers={'Date': 'Sun, 30 Oct 2016 03:13:52 GMT'})
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
@@ -1188,23 +1414,155 @@ def test_custom_date_header(loop, test_client):
 
 
 @asyncio.coroutine
-def test_response_task(loop, test_client):
-
-    srv_resp = None
+def test_response_prepared_with_clone(loop, test_client):
 
     @asyncio.coroutine
     def handler(request):
-        nonlocal srv_resp
-        srv_resp = web.StreamResponse()
-        assert srv_resp.task is None
-        yield from srv_resp.prepare(request)
-        assert srv_resp.task is not None
-        return srv_resp
+        cloned = request.clone()
+        resp = web.StreamResponse()
+        yield from resp.prepare(cloned)
+        return resp
 
-    app = web.Application(loop=loop)
+    app = web.Application()
     app.router.add_get('/', handler)
     client = yield from test_client(app)
 
     resp = yield from client.get('/')
     assert 200 == resp.status
-    assert srv_resp.task is None
+
+
+@asyncio.coroutine
+def test_app_max_client_size(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        yield from request.post()
+        return web.Response(body=b'ok')
+
+    max_size = 1024**2
+    app = web.Application()
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+    data = {"long_string": max_size * 'x' + 'xxx'}
+    resp = yield from client.post('/', data=data)
+    assert 413 == resp.status
+    resp_text = yield from resp.text()
+    assert 'Request Entity Too Large' in resp_text
+
+
+@asyncio.coroutine
+def test_app_max_client_size_adjusted(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        yield from request.post()
+        return web.Response(body=b'ok')
+
+    default_max_size = 1024**2
+    custom_max_size = default_max_size * 2
+    app = web.Application(client_max_size=custom_max_size)
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+    data = {'long_string': default_max_size * 'x' + 'xxx'}
+    resp = yield from client.post('/', data=data)
+    assert 200 == resp.status
+    resp_text = yield from resp.text()
+    assert 'ok' == resp_text
+    too_large_data = {'log_string': custom_max_size * 'x' + "xxx"}
+    resp = yield from client.post('/', data=too_large_data)
+    assert 413 == resp.status
+    resp_text = yield from resp.text()
+    assert 'Request Entity Too Large' in resp_text
+
+
+@asyncio.coroutine
+def test_app_max_client_size_none(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        yield from request.post()
+        return web.Response(body=b'ok')
+
+    default_max_size = 1024**2
+    custom_max_size = None
+    app = web.Application(client_max_size=custom_max_size)
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+    data = {'long_string': default_max_size * 'x' + 'xxx'}
+    resp = yield from client.post('/', data=data)
+    assert 200 == resp.status
+    resp_text = yield from resp.text()
+    assert 'ok' == resp_text
+    too_large_data = {'log_string': default_max_size * 2 * 'x'}
+    resp = yield from client.post('/', data=too_large_data)
+    assert 200 == resp.status
+    resp_text = yield from resp.text()
+    assert resp_text == 'ok'
+
+
+@asyncio.coroutine
+def test_post_max_client_size(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        try:
+            yield from request.post()
+        except ValueError:
+            return web.HTTPOk()
+        return web.HTTPBadRequest()
+
+    app = web.Application(client_max_size=10)
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+
+    data = {"long_string": 1024 * 'x', 'file': io.BytesIO(b'test')}
+    resp = yield from client.post('/', data=data)
+
+    assert 200 == resp.status
+
+
+@asyncio.coroutine
+def test_post_max_client_size_for_file(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        try:
+            yield from request.post()
+        except ValueError:
+            return web.HTTPOk()
+        return web.HTTPBadRequest()
+
+    app = web.Application(client_max_size=2)
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+
+    data = {'file': io.BytesIO(b'test')}
+    resp = yield from client.post('/', data=data)
+
+    assert 200 == resp.status
+
+
+@asyncio.coroutine
+def test_response_with_bodypart(loop, test_client):
+
+    @asyncio.coroutine
+    def handler(request):
+        reader = yield from request.multipart()
+        part = yield from reader.next()
+        return web.Response(body=part)
+
+    app = web.Application(client_max_size=2)
+    app.router.add_post('/', handler)
+    client = yield from test_client(app)
+
+    data = {'file': io.BytesIO(b'test')}
+    resp = yield from client.post('/', data=data)
+
+    assert 200 == resp.status
+    body = yield from resp.read()
+    assert body == b'test'
+
+    disp = multipart.parse_content_disposition(
+        resp.headers['content-disposition'])
+    assert disp == ('attachment',
+                    {'name': 'file', 'filename': 'file', 'filename*': 'file'})
