@@ -1,28 +1,19 @@
-import os
 from unittest import mock
 
-from yarl import URL
-
 from aiohttp import hdrs, helpers
-from aiohttp.file_sender import FileSender
 from aiohttp.test_utils import make_mocked_coro, make_mocked_request
-
-
-def test_env_nosendfile():
-    with mock.patch.dict(os.environ, {'AIOHTTP_NOSENDFILE': '1'}):
-        file_sender = FileSender()
-        assert file_sender._sendfile == file_sender._sendfile_fallback
+from aiohttp.web_fileresponse import FileResponse, SendfilePayloadWriter
 
 
 def test_static_handle_eof(loop):
     fake_loop = mock.Mock()
-    with mock.patch('aiohttp.file_sender.os') as m_os:
+    with mock.patch('aiohttp.web_fileresponse.os') as m_os:
         out_fd = 30
         in_fd = 31
         fut = helpers.create_future(loop)
         m_os.sendfile.return_value = 0
-        file_sender = FileSender()
-        file_sender._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
+        writer = SendfilePayloadWriter(fake_loop, mock.Mock())
+        writer._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
         m_os.sendfile.assert_called_with(out_fd, in_fd, 0, 100)
         assert fut.done()
         assert fut.result() is None
@@ -32,17 +23,17 @@ def test_static_handle_eof(loop):
 
 def test_static_handle_again(loop):
     fake_loop = mock.Mock()
-    with mock.patch('aiohttp.file_sender.os') as m_os:
+    with mock.patch('aiohttp.web_fileresponse.os') as m_os:
         out_fd = 30
         in_fd = 31
         fut = helpers.create_future(loop)
         m_os.sendfile.side_effect = BlockingIOError()
-        file_sender = FileSender()
-        file_sender._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
+        writer = SendfilePayloadWriter(fake_loop, mock.Mock())
+        writer._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
         m_os.sendfile.assert_called_with(out_fd, in_fd, 0, 100)
         assert not fut.done()
         fake_loop.add_writer.assert_called_with(out_fd,
-                                                file_sender._sendfile_cb,
+                                                writer._sendfile_cb,
                                                 fut, out_fd, in_fd, 0, 100,
                                                 fake_loop, True)
         assert not fake_loop.remove_writer.called
@@ -50,14 +41,14 @@ def test_static_handle_again(loop):
 
 def test_static_handle_exception(loop):
     fake_loop = mock.Mock()
-    with mock.patch('aiohttp.file_sender.os') as m_os:
+    with mock.patch('aiohttp.web_fileresponse.os') as m_os:
         out_fd = 30
         in_fd = 31
         fut = helpers.create_future(loop)
         exc = OSError()
         m_os.sendfile.side_effect = exc
-        file_sender = FileSender()
-        file_sender._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
+        writer = SendfilePayloadWriter(fake_loop, mock.Mock())
+        writer._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
         m_os.sendfile.assert_called_with(out_fd, in_fd, 0, 100)
         assert fut.done()
         assert exc is fut.exception()
@@ -67,13 +58,13 @@ def test_static_handle_exception(loop):
 
 def test__sendfile_cb_return_on_cancelling(loop):
     fake_loop = mock.Mock()
-    with mock.patch('aiohttp.file_sender.os') as m_os:
+    with mock.patch('aiohttp.web_fileresponse.os') as m_os:
         out_fd = 30
         in_fd = 31
         fut = helpers.create_future(loop)
         fut.cancel()
-        file_sender = FileSender()
-        file_sender._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
+        writer = SendfilePayloadWriter(fake_loop, mock.Mock())
+        writer._sendfile_cb(fut, out_fd, in_fd, 0, 100, fake_loop, False)
         assert fut.done()
         assert not fake_loop.add_writer.called
         assert not fake_loop.remove_writer.called
@@ -82,7 +73,7 @@ def test__sendfile_cb_return_on_cancelling(loop):
 
 def test_using_gzip_if_header_present_and_file_available(loop):
     request = make_mocked_request(
-        'GET', URL('http://python.org/logo.png'), headers={
+        'GET', 'http://python.org/logo.png', headers={
             hdrs.ACCEPT_ENCODING: 'gzip'
         }
     )
@@ -98,10 +89,10 @@ def test_using_gzip_if_header_present_and_file_available(loop):
     filepath.open = mock.mock_open()
     filepath.with_name.return_value = gz_filepath
 
-    file_sender = FileSender()
+    file_sender = FileResponse(filepath)
     file_sender._sendfile = make_mocked_coro(None)
 
-    loop.run_until_complete(file_sender.send(request, filepath))
+    loop.run_until_complete(file_sender.prepare(request))
 
     assert not filepath.open.called
     assert gz_filepath.open.called
@@ -109,7 +100,7 @@ def test_using_gzip_if_header_present_and_file_available(loop):
 
 def test_gzip_if_header_not_present_and_file_available(loop):
     request = make_mocked_request(
-        'GET', URL('http://python.org/logo.png'), headers={
+        'GET', 'http://python.org/logo.png', headers={
         }
     )
 
@@ -124,10 +115,10 @@ def test_gzip_if_header_not_present_and_file_available(loop):
     filepath.stat.return_value = mock.MagicMock()
     filepath.stat.st_size = 1024
 
-    file_sender = FileSender()
+    file_sender = FileResponse(filepath)
     file_sender._sendfile = make_mocked_coro(None)
 
-    loop.run_until_complete(file_sender.send(request, filepath))
+    loop.run_until_complete(file_sender.prepare(request))
 
     assert filepath.open.called
     assert not gz_filepath.open.called
@@ -135,7 +126,7 @@ def test_gzip_if_header_not_present_and_file_available(loop):
 
 def test_gzip_if_header_not_present_and_file_not_available(loop):
     request = make_mocked_request(
-        'GET', URL('http://python.org/logo.png'), headers={
+        'GET', 'http://python.org/logo.png', headers={
         }
     )
 
@@ -150,10 +141,10 @@ def test_gzip_if_header_not_present_and_file_not_available(loop):
     filepath.stat.return_value = mock.MagicMock()
     filepath.stat.st_size = 1024
 
-    file_sender = FileSender()
+    file_sender = FileResponse(filepath)
     file_sender._sendfile = make_mocked_coro(None)
 
-    loop.run_until_complete(file_sender.send(request, filepath))
+    loop.run_until_complete(file_sender.prepare(request))
 
     assert filepath.open.called
     assert not gz_filepath.open.called
@@ -161,7 +152,7 @@ def test_gzip_if_header_not_present_and_file_not_available(loop):
 
 def test_gzip_if_header_present_and_file_not_available(loop):
     request = make_mocked_request(
-        'GET', URL('http://python.org/logo.png'), headers={
+        'GET', 'http://python.org/logo.png', headers={
             hdrs.ACCEPT_ENCODING: 'gzip'
         }
     )
@@ -177,10 +168,10 @@ def test_gzip_if_header_present_and_file_not_available(loop):
     filepath.stat.return_value = mock.MagicMock()
     filepath.stat.st_size = 1024
 
-    file_sender = FileSender()
+    file_sender = FileResponse(filepath)
     file_sender._sendfile = make_mocked_coro(None)
 
-    loop.run_until_complete(file_sender.send(request, filepath))
+    loop.run_until_complete(file_sender.prepare(request))
 
     assert filepath.open.called
     assert not gz_filepath.open.called
