@@ -1,7 +1,9 @@
 import asyncio
 from unittest import mock
 
-from aiohttp import web
+import pytest
+
+from aiohttp import client, web
 
 
 @asyncio.coroutine
@@ -11,8 +13,8 @@ def test_simple_server(raw_test_server, test_client):
         return web.Response(text=str(request.rel_url))
 
     server = yield from raw_test_server(handler)
-    client = yield from test_client(server)
-    resp = yield from client.get('/path/to')
+    cli = yield from test_client(server)
+    resp = yield from cli.get('/path/to')
     assert resp.status == 200
     txt = yield from resp.text()
     assert txt == '/path/to'
@@ -28,8 +30,8 @@ def test_raw_server_not_http_exception(raw_test_server, test_client):
 
     logger = mock.Mock()
     server = yield from raw_test_server(handler, logger=logger)
-    client = yield from test_client(server)
-    resp = yield from client.get('/path/to')
+    cli = yield from test_client(server)
+    resp = yield from cli.get('/path/to')
     assert resp.status == 500
 
     txt = yield from resp.text()
@@ -40,6 +42,46 @@ def test_raw_server_not_http_exception(raw_test_server, test_client):
         exc_info=exc)
 
 
+@asyncio.coroutine
+def test_raw_server_handler_timeout(raw_test_server, test_client):
+    exc = asyncio.TimeoutError("error")
+
+    @asyncio.coroutine
+    def handler(request):
+        raise exc
+
+    logger = mock.Mock()
+    server = yield from raw_test_server(handler, logger=logger)
+    cli = yield from test_client(server)
+    resp = yield from cli.get('/path/to')
+    assert resp.status == 504
+
+    yield from resp.text()
+    logger.debug.assert_called_with("Request handler timed out.")
+
+
+@asyncio.coroutine
+def test_raw_server_do_not_swallow_exceptions(raw_test_server, test_client):
+    exc = None
+
+    @asyncio.coroutine
+    def handler(request):
+        raise exc
+
+    logger = mock.Mock()
+    server = yield from raw_test_server(handler, logger=logger)
+    cli = yield from test_client(server)
+
+    for _exc, msg in (
+            (asyncio.CancelledError("error"),
+             'Ignored premature client disconnection'),):
+        exc = _exc
+        with pytest.raises(client.ServerDisconnectedError):
+            yield from cli.get('/path/to')
+
+        logger.debug.assert_called_with(msg)
+
+
 @asyncio.coroutine
 def test_raw_server_not_http_exception_debug(raw_test_server, test_client):
     exc = RuntimeError("custom runtime error")
@@ -50,8 +92,8 @@ def test_raw_server_not_http_exception_debug(raw_test_server, test_client):
 
     logger = mock.Mock()
     server = yield from raw_test_server(handler, logger=logger, debug=True)
-    client = yield from test_client(server)
-    resp = yield from client.get('/path/to')
+    cli = yield from test_client(server)
+    resp = yield from cli.get('/path/to')
     assert resp.status == 500
 
     txt = yield from resp.text()
