@@ -10,6 +10,7 @@ import pytest
 from aiohttp import helpers
 from aiohttp.test_utils import make_mocked_coro
 
+
 base_worker = pytest.importorskip('aiohttp.worker')
 
 
@@ -23,6 +24,12 @@ WRONG_LOG_FORMAT = '%a "%{Referrer}i" %(h)s %(l)s %s'
 ACCEPTABLE_LOG_FORMAT = '%a "%{Referrer}i" %s'
 
 
+# tokio event loop does not allow to override attributes
+def skip_if_no_dict(loop):
+    if not hasattr(loop, '__dict__'):
+        pytest.skip("can not override loop attributes")
+
+
 class BaseTestWorker:
 
     def __init__(self):
@@ -31,6 +38,11 @@ class BaseTestWorker:
         self.cfg = mock.Mock()
         self.cfg.graceful_timeout = 100
 
+        try:
+            self.pid = 'pid'
+        except:
+            pass
+
 
 class AsyncioWorker(BaseTestWorker, base_worker.GunicornWebWorker):
     pass
@@ -152,15 +164,14 @@ def test_make_handler(worker, mocker):
 def test_make_handler_wsgi(worker, mocker):
     worker.wsgi = lambda env, start_resp: start_resp()
     worker.loop = mock.Mock()
+    worker.loop.time.return_value = 1477797232
     worker.log = mock.Mock()
     worker.cfg = mock.Mock()
     worker.cfg.access_log_format = ACCEPTABLE_LOG_FORMAT
     mocker.spy(worker, '_get_valid_log_format')
 
-    f = worker.make_handler(worker.wsgi)
-    assert isinstance(f, base_worker.WSGIServer)
-    assert isinstance(f(), base_worker.WSGIServerHttpProtocol)
-    assert worker._get_valid_log_format.called
+    with pytest.raises(RuntimeError):
+        worker.make_handler(worker.wsgi)
 
 
 @pytest.mark.parametrize('source,result', [
@@ -180,6 +191,8 @@ def test__get_valid_log_format_exc(worker):
 
 @asyncio.coroutine
 def test__run_ok(worker, loop):
+    skip_if_no_dict(loop)
+
     worker.ppid = 1
     worker.alive = True
     worker.servers = {}
@@ -217,6 +230,8 @@ def test__run_ok(worker, loop):
                     reason="UNIX sockets are not supported")
 @asyncio.coroutine
 def test__run_ok_unix_socket(worker, loop):
+    skip_if_no_dict(loop)
+
     worker.ppid = 1
     worker.alive = True
     worker.servers = {}
@@ -333,6 +348,8 @@ def test_close_wsgi(worker, loop):
 
 @asyncio.coroutine
 def test__run_ok_no_max_requests(worker, loop):
+    skip_if_no_dict(loop)
+
     worker.ppid = 1
     worker.alive = True
     worker.servers = {}
@@ -368,6 +385,8 @@ def test__run_ok_no_max_requests(worker, loop):
 
 @asyncio.coroutine
 def test__run_ok_max_requests_exceeded(worker, loop):
+    skip_if_no_dict(loop)
+
     worker.ppid = 1
     worker.alive = True
     worker.servers = {}
